import pandas as pd
import numpy as np
import datetime
from os import listdir
from os.path import exists
import yfinance as yf
np.random.seed(0)

class Data:
	def __init__(self, trading_year):
		wikidata = pd.read_html("https://en.wikipedia.org/wiki/List_of_S%26P_500_companies")[0]
		self.trading_year = trading_year
		self.sp500 = list(wikidata["Symbol"])
		self.START = str(trading_year-3)+"-01-01"
		self.END = str(trading_year+1)+"-01-06"
		print("imported data from {} to {}".format(self.START, self.END))
    
	def import_data(self, download_new_data = False):
		datafile_name = "data_{}-{}".format(self.trading_year-3,self.trading_year)
		if exists(datafile_name) and not download_new_data:
			print("{} is available already".format(datafile_name))
			df = pd.read_csv(datafile_name)
			return df
		self.n_companies = 0
		df = pd.DataFrame([], columns = ["Date"])
		for i, name in enumerate(self.sp500):
			cols = []
			try:
				ticker = yf.Ticker(name)
				hist = ticker.history(start = self.START, end = self.END).reset_index()
				if hist.shape[0] > 0:
					if self.n_companies == 0:
						df["Date"] = hist["Date"]
					df[name] = hist["Close"]
					self.n_companies += 1
					print("#",i, name, "was successfully appended!!!")
				else:
					print("#",i, name, "has not enough data")
			except:
				print("#",i, name, "failed to append...")
		df.to_csv(datafile_name, index = False, header = True)
		print("data has been saved with name {}".format(datafile_name))
		return df

	def number_of_companies(self):
		return self.n_companies

class Prepared_Data:
	def __init__(self, df, year):
		self.df = df
		try:
			self.df["Date"] = self.df["Date"].apply(lambda x: x.strftime("%Y-%m-%d"))
		except:
			print("Date is already string") 
		self.test_year = year

	def datasets(self):
		def create_label(df,perc=[0.5,0.5]):
			perc = [0.]+list(np.cumsum(perc))
			label = df.iloc[:,1:].pct_change(fill_method=None)[1:].apply(lambda x: pd.qcut(x.rank(method='first'),perc,labels=False), axis=1)
			return label
        
		def create_stock_data(df, st):
			st_data = pd.DataFrame([])
			st_data['Date'] = list(df['Date'])
			st_data['Name'] = [st]*len(st_data)
			for k in list(range(1,21))+list(range(40,241,20)):
				st_data['R'+str(k)] = df[st].pct_change(k)
			st_data['R-future'] = df[st].pct_change().shift(-1)  
			st_data['label'] = list(label[st])+[np.nan] 
			st_data['Month'] = list(df['Date'].str[:-3])
			st_data = st_data.dropna()
		    
			trade_year = st_data['Month'].str[:4]
			st_data = st_data.drop(columns=['Month'])
			st_train_data = st_data[trade_year<str(self.test_year)]
			st_test_data = st_data[trade_year==str(self.test_year)]
			return np.array(st_train_data),np.array(st_test_data)

		label = create_label(self.df)
		stock_names = sorted(list(self.df.columns[1:]))
		train_data,test_data = [],[]

		for st in stock_names:
			st_train_data,st_test_data = create_stock_data(self.df,st)
			train_data.append(st_train_data)
			test_data.append(st_test_data)

		train_data = np.concatenate([x for x in train_data])
		test_data = np.concatenate([x for x in test_data])

		self.y_training = train_data[:,-1]
		self.y_testing = test_data[:,-1]
		self.X_training = train_data[:,2:-2]
		self.X_testing = test_data[:,2:-2]

		return train_data, test_data

	def train_test_datasets(self):
		return self.X_training.astype(float), self.y_training.astype(float), self.X_testing.astype(float), self.y_testing.astype(float)


class ObtainData:
	def __init__(self, trading_year):
		self.finance_data = Data(trading_year = trading_year)
		self.data = self.finance_data.import_data()
		self.prepared_data = Prepared_Data(self.data, year = 2019)
		self.train_data, self.trading_data = self.prepared_data.datasets()
		self.X_training, self.y_training, self.X_testing, self.y_testing = self.prepared_data.train_test_datasets()
	
	def getSplitData(self):
		print("Train dataset size: ", self.train_data.shape, "Trading dataset size: ", self.trading_data.shape)
		print(self.train_data[:5,:3])
		print(self.trading_data[:5,:3])
		return self.X_training, self.y_training, self.X_testing, self.y_testing

	def getTrainTestData(self):
		return self.train_data, self.trading_data




import torch 
import torch.nn as nn
import torchvision.transforms as transforms
import torchvision.datasets as dsets
import torch.nn.functional as F
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import sys
from torch.utils.data import Dataset, DataLoader
from os import listdir
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.metrics import r2_score
from joblib import dump, load
from StatsArbData import ObtainData
torch.manual_seed(0)

class LoadData(Dataset):
    def __init__(self, X, y):
        self.x = torch.tensor(X).float()
        self.y = torch.tensor(y).long()
        self.len = self.x.shape[0]

    def __getitem__(self,index):
        return self.x[index],self.y[index]

    def __len__(self):
        return self.len

class DNN_1(nn.Module):
    def __init__(self, D_in, H1, H2, H3, D_out=2):
        super(DNN_1, self).__init__()
        self.drop1 = nn.Dropout(p = 0.1)
        self.drop2 = nn.Dropout(p = 0.5)
        self.channel1_l1 = nn.Linear(D_in, H1)
        self.channel1_l2 = nn.Linear(H1, H2)
        self.channel1_l3 = nn.Linear(H2, H3)
        
        self.channel2_l1 = nn.Linear(D_in, H1)
        self.channel2_l2 = nn.Linear(H1, H2)
        self.channel2_l3 = nn.Linear(H2, H3)
        
        self.out = nn.Linear(H3, D_out)

    def forward(self, x):
        x = torch.max(self.drop1(self.channel1_l1(x)), self.drop1(self.channel2_l1(x)))
        x = torch.max(self.drop2(self.channel1_l2(x)), self.drop2(self.channel2_l2(x)))
        x = torch.max(self.drop2(self.channel1_l3(x)), self.drop2(self.channel2_l3(x)))
        x = F.softmax(self.out(x))
        return x

class DNN_2(nn.Module):
    def __init__(self, D_in, H1, H2, H3, D_out=2):
        super(DNN_2, self).__init__()
        self.drop1 = nn.Dropout(p = 0.5)
        self.drop2 = nn.Dropout(p = 0.25)
        
        self.linear1_1 = nn.Linear(D_in, H1)
        torch.nn.init.kaiming_uniform_(self.linear1_1.weight, nonlinearity = "leaky_relu")
        self.linear2_1 = nn.Linear(H1, H2)
        torch.nn.init.kaiming_uniform_(self.linear2_1.weight, nonlinearity = "leaky_relu")
        self.linear3_1 = nn.Linear(H2, H3)
        torch.nn.init.kaiming_uniform_(self.linear3_1.weight, nonlinearity = "leaky_relu")
        self.linear1_2 = nn.Linear(D_in, H1)
        torch.nn.init.kaiming_uniform_(self.linear1_2.weight, nonlinearity = "leaky_relu")
        self.linear2_2 = nn.Linear(H1, H2)
        torch.nn.init.kaiming_uniform_(self.linear2_2.weight, nonlinearity = "leaky_relu")
        self.linear3_2 = nn.Linear(H2, H3)
        torch.nn.init.kaiming_uniform_(self.linear3_2.weight, nonlinearity = "leaky_relu")
        
        self.bn1 = nn.BatchNorm1d(H1)
        self.bn2 = nn.BatchNorm1d(H2)
        self.bn3 = nn.BatchNorm1d(H3)
        
        self.out = nn.Linear(H3, D_out)

    def forward(self, x):
        x = torch.max(self.bn1(self.drop1(self.linear1_1(x))), self.bn1(self.drop1(self.linear1_2(x))))
        x = torch.max(self.bn2(self.drop2(self.linear2_1(x))), self.bn2(self.drop2(self.linear2_2(x))))
        x = torch.max(self.bn3(self.linear3_1(x)), self.bn3(self.linear3_2(x)))
        x = F.softmax(self.out(x))
        return x


class ML_Models:
  def __init__(self, X_training, X_testing, y_training, y_testing):
      torch.manual_seed(0)
      self.deep_nn_1 = DNN_1(31, 31, 10, 5, 2)
      self.deep_nn_2 = DNN_2(31, 31, 10, 5, 2)
      self.random_forest = RandomForestClassifier(n_estimators = 1000, max_depth=20, max_features = "sqrt", random_state=0)
      self.gb_trees = GradientBoostingClassifier(n_estimators = 100, max_depth = 3, learning_rate = 0.1, min_samples_split = 15, random_state = 0)
      self.training = LoadData(X_training, y_training)
      self.testing = LoadData(X_testing, y_testing)
      self.X_training = X_training
      self.X_testing = X_testing 
      self.y_training = y_training
      self.y_testing = y_testing

  def train_dnn(self, model, criterion, train_loader, validation_loader, optimizer, paper_model=False, epochs=100, show_training_info = True):
      training_info = {'training_loss':[], 'validation_accuracy': []}
      MIN = -1
      COUNTER = 0
      temp = []
      for epoch in range(epochs):
          for i, (x, y) in enumerate(train_loader):
              optimizer.zero_grad()
              y_hat = model(x)
              if paper_model:
                  lambda1 = 0.00001
                  all_params = torch.cat([b.view(-1) for b in model.parameters()])
                  l1_regularization = lambda1 * torch.norm(all_params, 1)
                  loss = criterion(y_hat,y) + l1_regularization
              else:
                  loss = criterion(y_hat,y)
              loss.backward()
              optimizer.step()
              training_info['training_loss'].append(loss.data.item())
            
          if paper_model:
              if epoch < 400 or loss.data.item() < MIN:
                  MIN = loss.data.item()
                  COUNTER = 0
              else:
                  COUNTER += 1
                
          correct = 0
          accuracy = 0
          for x, y in validation_loader:
              z = model(x)
              _, y_hat = torch.max(z,1)
              correct = (y_hat == y).sum().item()
              accuracy += correct/x.shape[0]
          training_info['validation_accuracy'].append(accuracy/len(validation_loader))
          
          if show_training_info:
              print("LOSS:",training_info['training_loss'][-1],"ACCURCY:",training_info['validation_accuracy'][-1])
              print("current epoch",epoch+1,", ", 100*(epoch+1)/epochs, "% completed")

          if COUNTER == 5:
              print("FINISHED!!!")
              break
            
      return training_info

  def train_models(self, plot_graphs = False):
      # Define loss function and get training and testing datasets
      criterion = nn.CrossEntropyLoss()

      # Training DNN's paper model
      torch.manual_seed(0)
      train_loader = DataLoader(dataset = self.training, batch_size=2000, shuffle = True)
      validation_loader = DataLoader(dataset = self.testing, batch_size=5000)
      optimizer = torch.optim.Adadelta(self.deep_nn_1.parameters())


      path_1 = "dnn1.pth"
      try:
          self.deep_nn_1.load_state_dict(torch.load(path_1))
          self.deep_nn_1.eval()
          print("DNN 1 Model loaded")
      except:
          print("Training DNN Model 1...")
          dnn_1_info = self.train_dnn(self.deep_nn_1, criterion, train_loader, validation_loader, optimizer, paper_model = True, epochs = 500)
          if plot_graphs:
              steps = int(len(dnn_1_info["training_loss"])/len(dnn_1_info["validation_accuracy"]))
              plt.plot(dnn_1_info["training_loss"][::steps], color = "red")
              plt.plot(dnn_1_info["validation_accuracy"], color = "blue")
              plt.show()
          
          torch.save(self.deep_nn_1.state_dict(), path_1)
          print("DNN - Model 1 Accuracy: ", dnn_1_info["validation_accuracy"][-1])

      # Training DNN's our proposed model
      torch.manual_seed(0)
      train_loader = DataLoader(dataset = self.training, batch_size=2000, shuffle = True)
      validation_loader = DataLoader(dataset = self.testing, batch_size=5000)
      optimizer = torch.optim.Adam(self.deep_nn_2.parameters(), lr=1e-4)

      path_2 = "dnn2.pth"

      try:
          self.deep_nn_2.load_state_dict(torch.load(path_2))
          self.deep_nn_2.eval()
          print("DNN 2 Model loaded")
      except:
          print("Training DNN Model 2...")
          dnn_2_info = self.train_dnn(self.deep_nn_2, criterion, train_loader, validation_loader, optimizer, paper_model = False, epochs = 400)
          if plot_graphs:
              steps = int(len(dnn_2_info["training_loss"])/len(dnn_2_info["validation_accuracy"]))
              plt.plot(dnn_2_info["training_loss"][::steps], color = "red")
              plt.plot(dnn_2_info["validation_accuracy"], color = "blue")
              plt.show()

          torch.save(self.deep_nn_2.state_dict(), path_2)
          print("DNN- Model 2 Accuracy: ", dnn_2_info["validation_accuracy"][-1])
      
      

      # Training Random Forest
      try:
          self.random_forest = load("random_forest.joblib")
          print("Random Forest model loaded")
      except:
          self.random_forest.fit(self.X_training, self.y_training)
          dump(self.random_forest, "random_forest.joblib")
      print("Random Forest Accuracy:", self.random_forest.score(self.X_testing, self.y_testing))

      # Training Gradient Boosted Trees
      try:
          self.gb_trees = load("gb_tree.joblib")
          print("Gradient Boosted Trees model loaded")
      except:
          self.gb_trees.fit(self.X_training, self.y_training)
          dump(self.gb_trees, "gb_tree.joblib")
      print("Gradient Boosted Trees Accuracy:", self.gb_trees.score(self.X_testing, self.y_testing))
      print("TRAINING IS DONE!")

      return self.deep_nn_1, self.deep_nn_2, self.random_forest, self.gb_trees

  def import_DNN_1(self):
    return self.deep_nn_1
  
  def import_DNN_2(self):
    return self.deep_nn_2

  def import_rf(self):
    return self.random_forest
  
  def import_gbt(self):
    return self.gb_trees

class Models:
  def __init__(self, trading_year):
    self.dataset = ObtainData(trading_year = trading_year)
    self.X_training, self.y_training, self.X_testing, self.y_testing = self.dataset.getSplitData()
    self.models = ML_Models(self.X_training, self.X_testing, self.y_training, self.y_testing)	

  def getModels(self):
    self.dnn1, self.dnn2, self.rf, self.gbt = self.models.train_models(plot_graphs=False)
    return self.dnn1, self.dnn2, self.rf, self.gbt 



import pathlib
from setuptools import setup, find_packages
# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / "README.md").read_text()
# This call to setup() does all the work
setup(
	name="StatsArbPortfolio",
	version="1.0.1",
	description="StatsArbML",
	long_description=README,
	long_description_content_type="text/markdown",
	author="Carlos Gustavo Salas Flores",
	author_email="cs582@duke.edu",
	license="MIT",
	classifiers=[
	"License :: OSI Approved :: MIT License",
	"Programming Language :: Python :: 3",
	"Programming Language :: Python :: 3.7",
	],
	packages=find_packages(),
	# dependencies for installing the package
	install_requires=["numpy", "pandas", "torch", "sklearn", "yfinance", "matplotlib"]
)



import pandas as pd
import numpy as np
import datetime
from os import listdir
import yfinance as yf
import torch 
import torch.nn as nn
import torchvision.transforms as transforms
import torchvision.datasets as dsets
import torch.nn.functional as F
import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import sys
from torch.utils.data import Dataset, DataLoader
from os import listdir
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.metrics import r2_score
from joblib import dump, load
from StatsArbModels import Models
from StatsArbData import ObtainData

class Portfolio:
	def __init__(self, trading_year, stock):
		self.stock = stock
		temp = ObtainData(trading_year = trading_year)
		self.train_data, self.trading_data = temp.getTrainTestData()
		mask = pd.Series(self.trading_data[:,1]).apply(lambda x: True if x.upper() in [stock] else False)
		self.df = self.trading_data[mask]
		self.start_date = str(trading_year) + "-01-01"
		self.end_date = str(trading_year+1) + "-01-01"

	def simulation(self, name, prices, signal, init_budget = 10000):
		sharpe_ratio = [0]
		roi = [0]
		holdings = []
		cash = []
		total = []

		bnh_roi = [0]
		daily_returns = [0]
		bnh_sharpe_ratio = []

		initial_shares = init_budget // prices[0]

		name = name.upper()
		budget = init_budget
		shares = 0
		investment = 0
		length = prices.shape[0]
		
		for i, (curPrice, curSignal) in enumerate(zip(list(prices), list(signal))):
			if curSignal == 1 and shares == 0 and i < length-1:
            			shares = (budget // curPrice)
            			investment += shares * curPrice
            			budget -= investment
			elif investment > 0 and curSignal == 0 or i == length-1:
            			budget += shares * curPrice
            			shares = 0
            			investment = 0
            			roi.append(budget/init_budget - 1)
            			sharpe_ratio.append(np.mean(roi)/np.std(roi))
			holdings.append(shares*curPrice)
			cash.append(budget)
			total.append(budget + shares*curPrice)

			if i > 0:
				daily_returns.append((prices[i-1]-prices[i])/prices[i-1])
				bnh_roi.append((initial_shares*prices[i] - init_budget)/init_budget)

			if len(roi) < len(total):
				if i != 0:
					roi.append(roi[-1])
					sharpe_ratio.append(sharpe_ratio[-1])

			bnh_sharpe_ratio.append(np.mean(bnh_roi)/np.std(bnh_roi))
          
		print(name)
		print("final budget: $", budget)
		print("Algorithm Return: ", roi[-1])
		print("Algorithm Sharpe Ratio: ", sharpe_ratio[-1])

		bnh_return = (initial_shares*prices[-1]-init_budget)/init_budget
		print("Buy & Hold Return: ", bnh_return)
		print("Buy & Hold Sharpe Ratio: ", bnh_sharpe_ratio[-1])

		return roi[-1], bnh_return, (bnh_roi, bnh_sharpe_ratio, roi, sharpe_ratio, holdings, cash, total)

	def plot_trading(self, stk, model_name, prices, signals, dates, roi):
		dummy = pd.DataFrame([])
		dummy["Date"] = dates
		dummy["Price"] = prices
		dummy["Signal"] = signals
          
		mask = dummy["Signal"].diff() != 0
		mask[0] = True
      
		dummy["Signal"] = dummy["Signal"][mask]
		dummy = dummy.set_index("Date")
      
		plt.figure(figsize = (14,7))
		plt.title(stk + " " + model_name +" Signals. ROI: " + str(roi), fontsize = 8)
		plt.plot(dummy["Price"], color = "gray", label = "Closing Price")
		plt.plot(dummy["Price"][dummy["Signal"] == 1], '^', color = 'g', label = "BUY/HOLD")
		plt.plot(dummy["Price"][dummy["Signal"] == 0], 'v', color = 'r', label = "SELL/SHORT")
		plt.legend()

		plt.xticks(dummy.index[::15], fontsize = 16, rotation = 45)
		plt.yticks(fontsize = 16)
		plt.show()
  
	def plot_roi(self, stk, model_name, df):
		plt.figure(figsize = (14,7))
		plt.title("ROI for {} using {}".format(stk, model_name))
		plt.plot(df["ROI"], color = "b", label = "Algorithm")
		plt.plot(df["ROI BnH"], color = "r", label = "Buy & Hold")
		plt.legend()
	      
		plt.ylabel("ROI %")
		plt.show()

	def plot_sharpe_ratio(self, stk, model_name, df):
		plt.figure(figsize = (14,7))
		plt.title("Sharpe Ratio for {} using {}".format(stk, model_name))
		plt.plot(df["Sharpe Ratio"], color = "b", label="Algorithm")
		plt.plot(df["Sharpe Ratio BnH"], color = "r", label="Buy & Hold")
		plt.legend()

		plt.ylabel("Sharpe Ratio")
		plt.show()
  
	def plot_cash(self, stk, model_name, df):
		plt.figure(figsize = (14,7))
		plt.title("Returns for {} using {}".format(stk, model_name))
		plt.plot(df["Holdings"], color = "b", label = "Holdings")
		plt.plot(df["Cash"], color = "g", label = "Cash")
		plt.plot(df["Total"], color = "r", label = "Total")
		plt.legend()
		
		plt.ylabel("US$")
		plt.show()

	def BackTesting(self, model = None, model_name = "None", my_df = None, plot_holding_cash = True, ):
		name = self.stock.upper()
		stk = yf.Ticker(name)
		hist_data = stk.history(start = self.start_date, end = self.end_date).reset_index()
		prices = np.asanyarray(hist_data["Close"])
		dates = hist_data["Date"]

		X = self.trading_data[self.trading_data[:,1] == name][:,2:-2]

		signals = None
		if "DNN" in model_name:
			X = torch.tensor(X.astype(float)).float()
			z = model(X)
			_, y_hat = torch.max(z,1)
			signals = y_hat.numpy()
		else:
			signals = model.predict(X)

		roi, bnh_roi, (bnh_returns, bnh_sharpe_ratio, returns, sharpe_ratio, holdings, cash, total) = self.simulation(name, prices, signals)
		self.plot_trading(name, model_name, prices, signals, dates, roi)

		temp_df = pd.DataFrame([])
		temp_df["Date"] = dates
		temp_df["Sharpe Ratio BnH"] = np.array(bnh_sharpe_ratio)
		temp_df["Sharpe Ratio"] = np.array(sharpe_ratio)
		temp_df["ROI"] = np.array(returns)
		temp_df["ROI BnH"] = np.array(bnh_returns)
		temp_df["Holdings"] = np.array(holdings)
		temp_df["Cash"] = np.array(cash)
		temp_df["Total"] = np.array(total)
		temp_df = temp_df.set_index("Date")

		self.plot_cash(name, model_name, temp_df)
		self.plot_roi(name, model_name, temp_df)
		self.plot_sharpe_ratio(name, model_name, temp_df)

	def portfolios(self, model, model_id = 1):
		assert model_id == 1 or model_id == 2 or model_id == 3 or model_id == 4, "model_id should be 1 (DNN Model 1), 2 (DNN Model 2), 3 (Random Forest), or 4 (Gradient Boosted Trees)"
		if model_id == 1:
			self.BackTesting(model, "DNN Model 1")
		if model_id == 2:
			self.BackTesting(model, "DNN Model 2")
		if model_id == 3:
			self.BackTesting(model, "Random Forests")
		if model_id == 4:
			self.BackTesting(model, "Gradient-Boosted Trees")

class Simulation:
	def __init__(self, trading_year, stock, model_id):
		self.trading_year = trading_year
		self.portfolio = Portfolio(trading_year = trading_year, stock = stock)
		self.model_id = model_id
		self.my_model = Models(self.trading_year)
		self.dnn1, self.dnn2, self.rf, self.gbt = self.my_model.getModels()

	def plot_data(self):
		assert self.model_id == 1 or self.model_id == 2 or self.model_id == 3 or self.model_id == 4, "model_id is {} should be 1 (DNN Model 1), 2 (DNN Model 2), 3 (Random Forest), or 4 (Gradient Boosted Trees)".format(self.model_id)
		if self.model_id == 1:
			self.portfolio.portfolios(self.dnn1, model_id = self.model_id)
		if self.model_id == 2:
			self.portfolio.portfolios(self.dnn2, model_id = self.model_id)
		if self.model_id == 3:
			self.portfolio.portfolios(self.rf, model_id = self.model_id)
		if self.model_id == 4:
			self.portfolio.portfolios(self.gbt, model_id = self.model_id)
      
